@@ -1,8 +1,6 @@
 /*
   +----------------------------------------------------------------------+
-  | PHP Version 7                                                        |
-  +----------------------------------------------------------------------+
-  | Copyright (c) 2019 The PHP Group                                     |
+  | Copyright (c) The PHP Group                                          |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.01 of the PHP license,      |
   | that is bundled with this package in the file LICENSE, and is        |
@@ -199,33 +197,11 @@ static zend_always_inline zend_bool php_pcov_ignored_opcode(zend_uchar opcode) {
 } /* }}} */
 
 static zend_always_inline zend_string* php_pcov_interned_string(zend_string *string) { /* {{{ */
-	zend_string *interned;
-
 	if (ZSTR_IS_INTERNED(string)) {
 		return string;
 	}
 
-	if ((interned = zend_hash_find_ptr(&PCG(filenames), string))) {
-		return interned;
-	}
-
-	interned = (zend_string*) calloc(1, _ZSTR_STRUCT_SIZE(ZSTR_LEN(string)));
-
-	memcpy(
-		interned,
-		string,
-		_ZSTR_STRUCT_SIZE(ZSTR_LEN(string)));
-
-	GC_SET_REFCOUNT(interned, 1);
-#if PHP_VERSION_ID < 70200
-    GC_FLAGS(interned) |= IS_STR_PERMANENT | IS_STR_INTERNED;
-#else
-	GC_TYPE_INFO(interned) =
-			IS_STRING |
-			((IS_STR_INTERNED | IS_STR_PERMANENT) << GC_FLAGS_SHIFT);
-#endif
-
-	return zend_hash_add_ptr(&PCG(filenames), interned, interned);
+	return zend_new_interned_string(zend_string_copy(string));
 } /* }}} */
 
 static zend_always_inline php_coverage_t* php_pcov_create(zend_execute_data *execute_data) { /* {{{ */
@@ -240,9 +216,35 @@ static zend_always_inline php_coverage_t* php_pcov_create(zend_execute_data *exe
 	return create;
 } /* }}} */
 
+static zend_always_inline int php_pcov_has(zend_string *filename, uint32_t lineno) { /* {{{ */
+	HashTable *table = zend_hash_find_ptr(&PCG(covered), filename);
+
+	if (UNEXPECTED(!table)) {
+		HashTable covering;
+
+		zend_hash_init(&covering, 64, NULL, NULL, 0);
+
+		table = zend_hash_add_mem(
+			&PCG(covered), filename, &covering, sizeof(HashTable));
+
+ 		zend_hash_index_add_empty_element(table, lineno);
+		return 0;
+	}
+
+	if (EXPECTED(zend_hash_index_exists(table, lineno))) {
+		return 1;
+	}
+
+	zend_hash_index_add_empty_element(table, lineno);
+	return 0;
+} /* }}} */
+
 static zend_always_inline int php_pcov_trace(zend_execute_data *execute_data) { /* {{{ */
-	if (PCG(enabled)) {
-		if (php_pcov_wants(EX(func)->op_array.filename) && !php_pcov_ignored_opcode(EX(opline)->opcode)) {
+    if (PCG(enabled)) {
+		if (php_pcov_wants(EX(func)->op_array.filename) &&
+			!php_pcov_ignored_opcode(EX(opline)->opcode) &&
+			!php_pcov_has(EX(func)->op_array.filename, EX(opline)->lineno)) {
+
 			php_coverage_t *coverage = php_pcov_create(execute_data);
 
 			if (!PCG(start)) {
@@ -293,7 +295,7 @@ zend_op_array* php_pcov_compile_file(zend_file_handle *fh, int type) { /* {{{ */
 } /* }}} */
 
 void php_pcov_execute_ex(zend_execute_data *execute_data) { /* {{{ */
-	int zrc        = 0;
+	int zrc		= 0;
 
 	while (1) {
 		zrc = php_pcov_trace(execute_data);
@@ -307,6 +309,11 @@ void php_pcov_execute_ex(zend_execute_data *execute_data) { /* {{{ */
 	}
 } /* }}} */
 
+void php_pcov_covered_dtor(zval *zv) { /* {{{ */
+	zend_hash_destroy(Z_PTR_P(zv));
+	efree(Z_PTR_P(zv));
+} /* }}} */
+
 void php_pcov_files_dtor(zval *zv) { /* {{{ */
 	destroy_op_array(Z_PTR_P(zv));
 	efree(Z_PTR_P(zv));
@@ -425,7 +432,7 @@ PHP_RINIT_FUNCTION(pcov)
 	zend_hash_init(&PCG(ignores),    INI_INT("pcov.initial.files"), NULL, NULL, 0);
 	zend_hash_init(&PCG(wants),      INI_INT("pcov.initial.files"), NULL, NULL, 0);
 	zend_hash_init(&PCG(discovered), INI_INT("pcov.initial.files"), NULL, ZVAL_PTR_DTOR, 0);
-	zend_hash_init(&PCG(filenames),  INI_INT("pcov.initial.files"), NULL, php_pcov_filename_dtor, 0);
+	zend_hash_init(&PCG(covered),    INI_INT("pcov.initial.files"), NULL, php_pcov_covered_dtor, 0);
 
 	php_pcov_setup_directory(INI_STR("pcov.directory"));
 	php_pcov_setup_exclude(INI_STR("pcov.exclude"));
@@ -460,7 +467,7 @@ PHP_RSHUTDOWN_FUNCTION(pcov)
 	zend_hash_destroy(&PCG(wants));
 	zend_hash_destroy(&PCG(discovered));
 	zend_hash_destroy(&PCG(waiting));
-	zend_hash_destroy(&PCG(filenames));
+	zend_hash_destroy(&PCG(covered));
 
 	zend_arena_destroy(PCG(mem));
 
@@ -537,7 +544,7 @@ static zend_always_inline void php_pcov_report(php_coverage_t *coverage, zval *f
 	} while ((coverage = coverage->next));
 } /* }}} */
 
-static zend_always_inline void php_pcov_discover_code(zend_arena **arena, zend_op_array *ops, zval *return_value) { /* {{{ */
+static void php_pcov_discover_code(zend_arena **arena, zend_op_array *ops, zval *return_value) { /* {{{ */
 	zend_cfg cfg;
 	zend_basic_block *block;
 	zend_op *limit = ops->opcodes + ops->last;
@@ -592,6 +599,12 @@ static zend_always_inline void php_pcov_discover_code(zend_arena **arena, zend_o
 			break;
 		}
 	}
+
+#if PHP_VERSION_ID >= 80100
+    for (uint32_t def = 0; def < ops->num_dynamic_func_defs; def++) {
+        php_pcov_discover_code(arena, ops->dynamic_func_defs[def], return_value);
+    }
+#endif
 } /* }}} */
 
 static void php_pcov_discover_file(zend_string *file, zval *return_value) { /* {{{ */
@@ -782,6 +795,7 @@ PHP_NAMED_FUNCTION(php_pcov_clear)
 	PCG(next) = NULL;
 
 	php_pcov_clean(&PCG(waiting));
+	php_pcov_clean(&PCG(covered));
 } /* }}} */
 
 /* {{{ array \pcov\waiting(void) */
@@ -853,7 +867,7 @@ const zend_function_entry php_pcov_functions[] = {
 /* {{{ pcov_module_deps[] */
 static const zend_module_dep pcov_module_deps[] = {
 	ZEND_MOD_REQUIRED("pcre")
-	{NULL, NULL, NULL}
+	ZEND_MOD_END
 }; /* }}} */
 
 /* {{{ pcov_module_entry
